"""
Betty/pytest test utilities.

Add ``from betty.test_utils.conftest import *`` to your project's ``conftest.py``
to start using these utilities.
"""

from __future__ import annotations

__all__ = [
    "binary_file_cache",
    "multiprocessing_manager",
    "new_temporary_app",
    "new_temporary_app_factory",
    "page",
    "process_pool",
]

import multiprocessing
from typing import TYPE_CHECKING, Protocol, AsyncContextManager, Any

import pytest
import pytest_asyncio

from betty.app import App
from betty.cache.file import BinaryFileCache
from betty.error import do_raise
from betty.multiprocessing import ProcessPoolExecutor

if TYPE_CHECKING:
    from playwright.async_api import Page, BrowserContext
    from betty.service import ServiceFactory
    from betty.cache import Cache
    from betty.fetch import Fetcher
    from concurrent import futures
    from multiprocessing.managers import SyncManager
    from pathlib import Path
    from collections.abc import AsyncIterator, Iterator


@pytest.fixture
async def binary_file_cache(
    multiprocessing_manager: SyncManager, tmp_path: Path
) -> BinaryFileCache:
    """
    Create a temporary binary file cache.
    """
    return BinaryFileCache(tmp_path, manager=multiprocessing_manager)


@pytest.fixture(scope="session")
async def process_pool() -> AsyncIterator[futures.ProcessPoolExecutor]:
    """
    Create a new, temporary :py:class:`betty.app.App`.
    """
    with ProcessPoolExecutor() as process_pool:
        yield process_pool


@pytest.fixture(scope="session")
def multiprocessing_manager() -> Iterator[SyncManager]:
    """
    Raise Betty's own deprecation warnings as errors.
    """
    with multiprocessing.Manager() as manager:
        yield manager


@pytest.fixture
async def new_temporary_app(
    process_pool: futures.ProcessPoolExecutor, multiprocessing_manager: SyncManager
) -> AsyncIterator[App]:
    """
    Create a new, temporary :py:class:`betty.app.App`.
    """
    async with (
        App.new_temporary(
            process_pool=process_pool, multiprocessing_manager=multiprocessing_manager
        ) as app,
        app,
    ):
        yield app


class NewTemporaryAppFactory(Protocol):
    async def __call__(
        self,
        *,
        fetcher: Fetcher | None = None,
        process_pool: futures.ProcessPoolExecutor | None = None,
    ) -> AsyncContextManager[App]:
        pass


@pytest.fixture
async def new_temporary_app_factory(
    process_pool: futures.ProcessPoolExecutor, multiprocessing_manager: SyncManager
) -> NewTemporaryAppFactory:
    """
    Get a factory to create a new, temporary :py:class:`betty.app.App`.
    """
    fixture_process_pool = process_pool
    fixture_multiprocessing_manager = multiprocessing_manager

    async def _new_temporary_app_factory(
        *,
        cache_factory: ServiceFactory[App, Cache[Any]] | None = None,
        fetcher: Fetcher | None = None,
        process_pool: futures.ProcessPoolExecutor | None = None,
        multiprocessing_manager: SyncManager | None = None,
    ) -> AsyncContextManager[App]:
        return App.new_temporary(
            cache_factory=cache_factory,
            fetcher=fetcher,
            process_pool=process_pool or fixture_process_pool,
            multiprocessing_manager=multiprocessing_manager
            or fixture_multiprocessing_manager,
        )

    return _new_temporary_app_factory


@pytest_asyncio.fixture(loop_scope="session")
async def page(context: BrowserContext) -> Page:
    """
    A Playwright Page instance.
    """
    page = await context.new_page()
    page.on("pageerror", do_raise)
    return page



"""
Provide configuration for the :py:class:`betty.project.extension.gramps.Gramps` extension.
"""

from __future__ import annotations

from pathlib import Path
from typing import Any, final, TYPE_CHECKING, TypeVar
from warnings import warn

from typing_extensions import override

from betty.assertion import (
    RequiredField,
    OptionalField,
    assert_record,
    assert_path,
    assert_setattr,
    assert_mapping,
    assert_len,
    assert_str,
)
from betty.config import Configuration
from betty.config.collections.sequence import ConfigurationSequence
from betty.gramps.loader import (
    DEFAULT_EVENT_TYPES_MAPPING,
    DEFAULT_PLACE_TYPES_MAPPING,
    DEFAULT_PRESENCE_ROLES_MAPPING,
)
from betty.plugin import Plugin
from betty.plugin.config import PluginInstanceConfiguration
from betty.typing import internal

if TYPE_CHECKING:
    from betty.mutability import Mutable
    from betty.serde.dump import Dump, DumpMapping
    from collections.abc import Mapping, MutableMapping, Iterable, Iterator

_PluginT = TypeVar("_PluginT", bound=Plugin)


def _assert_gramps_type(value: Any) -> str:
    event_type = assert_str()(value)
    assert_len(minimum=1)(event_type)
    return event_type


@internal
@final
class PluginMapping(Configuration):
    """
    Map Gramps types to Betty plugin instances.
    """

    def __init__(
        self,
        default_mapping: Mapping[str, PluginInstanceConfiguration],
        mapping: Mapping[str, PluginInstanceConfiguration],
    ):
        super().__init__()
        self._default_mapping = default_mapping
        self._mapping: MutableMapping[str, PluginInstanceConfiguration] = {
            **default_mapping,
            **mapping,
        }

    @override
    def load(self, dump: Dump) -> None:
        self.assert_mutable()
        self._mapping = {
            **self._default_mapping,
            **assert_mapping(self._load_item, _assert_gramps_type)(dump),
        }

    def _load_item(self, dump: Dump) -> PluginInstanceConfiguration:
        configuration = PluginInstanceConfiguration("-")
        configuration.load(dump)
        return configuration

    @override
    def dump(self) -> Dump:
        return {
            gramps_type: configuration.dump()
            for gramps_type, configuration in self._mapping.items()
        }

    def __getitem__(self, gramps_type: str) -> PluginInstanceConfiguration:
        return self._mapping[gramps_type]

    def __setitem__(
        self, gramps_type: str, configuration: PluginInstanceConfiguration
    ) -> None:
        self.assert_mutable()
        self._mapping[gramps_type] = configuration

    def __delitem__(self, gramps_type: str) -> None:
        self.assert_mutable()
        del self._mapping[gramps_type]

    def __iter__(self) -> Iterator[str]:
        return iter(self._mapping)


class FamilyTreeConfiguration(Configuration):
    """
    Configure a single Gramps family tree.

    The ``genders`` argument has been deprecated since Betty 0.4.13. There is no alternative.
    """

    def __init__(
        self,
        file_path: Path,
        *,
        event_types: Mapping[str, PluginInstanceConfiguration] | None = None,
        place_types: Mapping[str, PluginInstanceConfiguration] | None = None,
        presence_roles: Mapping[str, PluginInstanceConfiguration] | None = None,
        genders: Mapping[str, PluginInstanceConfiguration] | None = None,
    ):
        super().__init__()
        self.file_path = file_path
        self._event_types = PluginMapping(
            {
                gramps_value: PluginInstanceConfiguration(event_type)
                for gramps_value, event_type in DEFAULT_EVENT_TYPES_MAPPING.items()
            },
            event_types or {},
        )
        if genders is not None:
            warn(
                "The ``genders`` argument has been deprecated since Betty 0.4.13. There is no alternative.",
                category=DeprecationWarning,
                stacklevel=2,
            )
        self._genders = PluginMapping({}, genders or {})
        self._place_types = PluginMapping(
            {
                gramps_value: PluginInstanceConfiguration(event_type)
                for gramps_value, event_type in DEFAULT_PLACE_TYPES_MAPPING.items()
            },
            place_types or {},
        )
        self._presence_roles = PluginMapping(
            {
                gramps_value: PluginInstanceConfiguration(event_type)
                for gramps_value, event_type in DEFAULT_PRESENCE_ROLES_MAPPING.items()
            },
            presence_roles or {},
        )

    @override
    def get_mutable_instances(self) -> Iterable[Mutable]:
        return (
            self._event_types,
            self._genders,
            self._place_types,
            self._presence_roles,
        )

    @property
    def file_path(self) -> Path | None:
        """
        The path to the Gramps family tree file.
        """
        return self._file_path

    @file_path.setter
    def file_path(self, file_path: Path | None) -> None:
        self.assert_mutable()
        self._file_path = file_path

    @property
    def event_types(self) -> PluginMapping:
        """
        How to map event types.
        """
        return self._event_types

    @property
    def genders(self) -> PluginMapping:
        """
        How to map genders.
        """
        return self._genders

    @property
    def place_types(self) -> PluginMapping:
        """
        How to map place types.
        """
        return self._place_types

    @property
    def presence_roles(self) -> PluginMapping:
        """
        How to map presence roles.
        """
        return self._presence_roles

    @override
    def load(self, dump: Dump) -> None:
        self.assert_mutable()
        assert_record(
            RequiredField("file", assert_path() | assert_setattr(self, "file_path")),
            OptionalField("event_types", self.event_types.load),
            OptionalField("genders", self.genders.load),
            OptionalField("place_types", self.place_types.load),
            OptionalField("presence_roles", self.presence_roles.load),
        )(dump)

    @override
    def dump(self) -> DumpMapping[Dump]:
        return {
            "file": str(self.file_path) if self.file_path else None,
            "event_types": self.event_types.dump(),
            "genders": self.genders.dump(),
            "place_types": self.place_types.dump(),
            "presence_roles": self.presence_roles.dump(),
        }


class FamilyTreeConfigurationSequence(ConfigurationSequence[FamilyTreeConfiguration]):
    """
    Configure zero or more Gramps family trees.
    """

    @override
    def _load_item(self, dump: Dump) -> FamilyTreeConfiguration:
        # Use a dummy path to satisfy initializer arguments.
        # It will be overridden when loading the dump.
        item = FamilyTreeConfiguration(Path())
        item.load(dump)
        return item


class GrampsConfiguration(Configuration):
    """
    Provide configuration for the :py:class:`betty.project.extension.gramps.Gramps` extension.
    """

    def __init__(
        self, *, family_trees: Iterable[FamilyTreeConfiguration] | None = None
    ):
        super().__init__()
        self._family_trees = FamilyTreeConfigurationSequence(family_trees)

    @override
    def get_mutable_instances(self) -> Iterable[Mutable]:
        return (self._family_trees,)

    @property
    def family_trees(self) -> FamilyTreeConfigurationSequence:
        """
        The Gramps family trees to load.
        """
        return self._family_trees

    @override
    def load(self, dump: Dump) -> None:
        self.assert_mutable()
        assert_record(OptionalField("family_trees", self.family_trees.load))(dump)

    @override
    def dump(self) -> DumpMapping[Dump]:
        return {"family_trees": self.family_trees.dump()}



"""Provide Betty's data model API."""

from __future__ import annotations

from abc import abstractmethod
from inspect import getmembers
from reprlib import recursive_repr
from typing import TypeVar, Any, Self, TypeAlias, TYPE_CHECKING, Iterable
from uuid import uuid4

import typing_extensions
from typing_extensions import override

from betty.json.linked_data import (
    LinkedDataDumpableJsonLdObject,
    JsonLdObject,
)
from betty.json.schema import JsonSchemaReference, Array, String, OneOf, Null
from betty.locale.localizable import (
    _,
    Localizable,
    StaticTranslationsLocalizableAttr,
    ngettext,
)
from betty.locale.localizer import DEFAULT_LOCALIZER
from betty.mutability import Mutable
from betty.plugin import PluginRepository, Plugin
from betty.plugin.entry_point import EntryPointPluginRepository
from betty.repr import repr_instance
from betty.string import kebab_case_to_lower_camel_case
from betty.user import UserFacing
from betty.warnings import deprecated, deprecate

if TYPE_CHECKING:
    from betty.serde.dump import DumpMapping, Dump
    from betty.project import Project
    import builtins

ENTITY_TYPE_REPOSITORY: PluginRepository[Entity] = EntryPointPluginRepository(
    "betty.entity_type"
)
"""
The entity type plugin repository.

Read more about :doc:`/development/plugin/entity-type`.
"""


class NonPersistentId(str):
    """
    A randomly generated ID that is not persistent.

    Entities must have IDs for identification. However, not all entities can be provided with an ID that exists in the
    original data set (such as a third-party family tree loaded into Betty).

    Non-persistent IDs are helpful in case there is no external ID that can be used. However, as they do not persist
    when reloading an ancestry, they *MUST NOT* be in contexts where persistent identifiers are expected, such as in
    URLs.
    """

    __slots__ = ()

    def __new__(cls, entity_id: str | None = None):  # noqa D102
        return super().__new__(cls, entity_id or str(uuid4()))


class Entity(LinkedDataDumpableJsonLdObject, Mutable, Plugin):
    """
    An entity is a uniquely identifiable data container.

    Read more about :doc:`/development/plugin/entity-type`.

    To test your own subclasses, use :py:class:`betty.test_utils.model.EntityTestBase`.
    """

    def __init__(
        self,
        id: str | None = None,  # noqa A002
        *args: Any,
        **kwargs: Any,
    ):
        self._id = NonPersistentId() if id is None else id
        super().__init__(*args, **kwargs)

    def __hash__(self) -> int:
        return hash(self.ancestry_id)

    @classmethod
    @abstractmethod
    def plugin_label_plural(cls) -> Localizable:
        """
        The human-readable entity type label, plural.
        """
        pass

    @classmethod
    def plugin_label_count(cls, count: int) -> Localizable:
        """
        The human-readable entity type label for the given entity count.
        """
        deprecate(
            f"Not overriding {cls}.plugin_label_count() has been deprecated since Betty 0.4.13."
        )
        return ngettext("{count} entity", "{count} entities", count).format(
            count=str(count)
        )

    @override  # type: ignore[callable-functiontype]
    @recursive_repr()
    def __repr__(self) -> str:
        return repr_instance(self, id=self._id)

    @property
    def type(self) -> builtins.type[Self]:
        """
        The entity type.
        """
        return self.__class__

    @property
    def id(self) -> str:
        """
        The entity ID.

        This MUST be unique per entity type, per ancestry.
        """
        return self._id

    @property
    def ancestry_id(self) -> tuple[builtins.type[Self], str]:
        """
        The ancestry ID.

        This MUST be unique per ancestry.
        """
        return self.type, self.id

    @property
    def label(self) -> Localizable:
        """
        The entity's human-readable label.
        """
        return _("{entity_type} {entity_id}").format(
            entity_type=self.plugin_label(), entity_id=self.id
        )

    @override
    async def dump_linked_data(self, project: Project) -> DumpMapping[Dump]:
        dump = await super().dump_linked_data(project)

        if persistent_id(self) and isinstance(self, UserFacing):
            url_generator = await project.url_generator
            dump["@id"] = url_generator.generate(
                f"betty-static:///{self.type.plugin_id()}/{self.id}/index.json",
                absolute=True,
            )
        dump["id"] = self.id

        return dump

    @override
    @classmethod
    async def linked_data_schema(cls, project: Project) -> JsonLdObject:
        schema = await super().linked_data_schema(project)
        schema._def_name = f"{kebab_case_to_lower_camel_case(cls.plugin_id())}Entity"
        schema.title = cls.plugin_label().localize(DEFAULT_LOCALIZER)
        schema.add_property("$schema", JsonSchemaReference())
        schema.add_property("id", String(title="Entity ID"), False)

        return schema

    @override
    def get_mutable_instances(self) -> Iterable[Mutable]:
        yield from super().get_mutable_instances()
        for __, member in getmembers(type(self)):
            if isinstance(member, StaticTranslationsLocalizableAttr):
                yield member.__get__(self, type(self))


AncestryEntityId: TypeAlias = tuple[type[Entity], str]


def persistent_id(entity_or_id: Entity | str) -> bool:
    """
    Test if an entity ID is persistent.

    See :py:class:`betty.model.NonPersistentId`.
    """
    return not isinstance(
        entity_or_id if isinstance(entity_or_id, str) else entity_or_id.id,
        NonPersistentId,
    )


@typing_extensions.deprecated(
    f"This class has been deprecated since Betty 0.4.13, and will be removed in Betty 0.5. Instead use {UserFacing}."
)
class UserFacingEntity(UserFacing):
    """
    A sentinel to mark an entity type as being visible to users (e.g. not internal).
    """

    pass


_EntityT = TypeVar("_EntityT", bound=Entity)


class ToZeroOrOneSchema(OneOf):
    """
    A schema for a to-zero-or-one entity association.
    """

    def __init__(self, *, title: str | None = None, description: str | None = None):
        super().__init__(
            String(
                title=title or "Optional associate entity",
                description=description
                or "An optional reference to an associate entity's JSON resource",
                format=String.Format.URI,
            ),
            Null(),
        )


class ToOneSchema(String):
    """
    A schema for a to-one entity association.
    """

    def __init__(self, *, title: str | None = None, description: str | None = None):
        super().__init__(
            title=title or "Associate entity",
            description=description
            or "A reference to an associate entity's JSON resource",
            format=String.Format.URI,
        )


@deprecated(
    f"This class has been deprecated since Betty 0.4.13, and will be removed in Betty 0.5. Instead use {ToOneSchema}."
)
class EntityReferenceSchema(ToOneSchema):
    """
    A schema for a reference to another entity resource.
    """

    pass


class ToManySchema(Array):
    """
    A schema for a to-many entity association.
    """

    def __init__(self, *, title: str | None = None, description: str | None = None):
        super().__init__(
            ToOneSchema(),
            title=title or "Associate entities",
            description=description
            or "References to associate entities' JSON resources",
        )


@deprecated(
    f"This class has been deprecated since Betty 0.4.13, and will be removed in Betty 0.5. Instead use {ToManySchema}."
)
class EntityReferenceCollectionSchema(ToManySchema):
    """
    A schema for a collection of references to other entity resources.
    """

    pass



"""
Provide the Cotton Candy theme.
"""

from __future__ import annotations

from pathlib import Path
from typing import TYPE_CHECKING, final

from typing_extensions import override, deprecated

from betty.jinja2 import Jinja2Provider, Filters
from betty.locale.localizable import _, static, plain
from betty.os import link_or_copy
from betty.plugin import ShorthandPluginBase
from betty.project.extension import ConfigurableExtension, Theme, Extension
from betty.project.extension._theme import jinja2_filters
from betty.project.extension._theme.search import generate_search_index
from betty.project.extension.cotton_candy.config import CottonCandyConfiguration
from betty.project.extension.maps import Maps
from betty.project.extension.trees import Trees
from betty.project.extension.webpack import Webpack
from betty.project.extension.webpack.build import EntryPointProvider
from betty.project.generate import GenerateSiteEvent

if TYPE_CHECKING:
    from betty.plugin import PluginIdentifier
    from betty.event_dispatcher import EventHandlerRegistry
    from collections.abc import Sequence

_RESULT_CONTAINER_TEMPLATE = plain("""
<li class="search-result">
    {{{ betty-search-result }}}
</li>
""")

_RESULTS_CONTAINER_TEMPLATE = plain("""
<ul id="search-results" class="nav-secondary">
    {{{ betty-search-results }}}
</ul>
""")


async def _generate_logo(event: GenerateSiteEvent) -> None:
    await link_or_copy(
        event.project.logo, event.project.configuration.www_directory_path / "logo.png"
    )


async def _generate_search_index(event: GenerateSiteEvent) -> None:
    await generate_search_index(
        event.project,
        _RESULT_CONTAINER_TEMPLATE,
        _RESULTS_CONTAINER_TEMPLATE,
        job_context=event.job_context,
    )


@final
@deprecated(
    "The Cotton Candy theme has been deprecated since Betty 0.4.9, and will be removed in Betty 0.5. Instead use Raspberry Mint (`raspberry-mint`)."
)
class CottonCandy(
    ShorthandPluginBase,
    Theme,
    ConfigurableExtension[CottonCandyConfiguration],
    Jinja2Provider,
    EntryPointProvider,
):
    """
    The Cotton Candy theme.
    """

    _plugin_id = "cotton-candy"
    _plugin_label = static("Cotton Candy")
    _plugin_description = _("Cotton Candy is Betty's legacy theme.")

    @override
    async def bootstrap(self) -> None:
        await super().bootstrap()
        try:
            await self._assert_configuration()
        except BaseException:
            await self.shutdown()
            raise

    async def _assert_configuration(self) -> None:
        await self.configuration.featured_entities.validate(
            self.project.entity_type_repository
        )

    @override
    def register_event_handlers(self, registry: EventHandlerRegistry) -> None:
        registry.add_handler(GenerateSiteEvent, _generate_logo, _generate_search_index)

    @override
    @classmethod
    def depends_on(cls) -> set[PluginIdentifier[Extension]]:
        return {Webpack}

    @override
    @classmethod
    def comes_before(cls) -> set[PluginIdentifier[Extension]]:
        return {Maps, Trees}

    @override
    @classmethod
    def assets_directory_path(cls) -> Path:
        return Path(__file__).parent / "assets"

    @override
    @classmethod
    def webpack_entry_point_directory_path(cls) -> Path:
        return Path(__file__).parent / "webpack"

    @override
    def webpack_entry_point_cache_keys(self) -> Sequence[str]:
        return (
            self.project.configuration.root_path,
            self._configuration.primary_inactive_color.hex,
            self._configuration.primary_active_color.hex,
            self._configuration.link_inactive_color.hex,
            self._configuration.link_active_color.hex,
        )

    @override
    @classmethod
    def new_default_configuration(cls) -> CottonCandyConfiguration:
        return CottonCandyConfiguration()

    @override
    @property
    def filters(self) -> Filters:
        return jinja2_filters(self._project)



"""
Integrate Webpack with Jinja2.
"""

from jinja2.runtime import Context


def _context_js_entry_points(context: Context) -> set[str]:
    entry_points = context.resolve_or_missing("webpack_js_entry_points")
    if isinstance(entry_points, set):
        return entry_points
    raise RuntimeError(
        "No `webpack_js_entry_points` context variable exists in this Jinja2 template."
    )



"""
The mutability API.

This provides tools to mark objects as mutable or immutable, and to guard against mutations.
"""

from __future__ import annotations

from typing import Any, TYPE_CHECKING

from betty.typing import internal

if TYPE_CHECKING:
    from collections.abc import Iterable


class MutabilityError(Exception):
    """
    A generic mutability API error.
    """

    pass


@internal
class MutableError(MutabilityError, RuntimeError):
    """
    An error raised because something was unexpectedly mutable.
    """

    pass


@internal
class ImmutableError(MutabilityError, RuntimeError):
    """
    An error raised because something was unexpectedly immutable.
    """

    pass


class Mutable:
    """
    A generic mutable type that can be marked immutable.
    """

    def __init__(self, *args: Any, mutable: bool = True, **kwargs: Any):
        super().__init__(*args, **kwargs)
        self._mutable = mutable

    def get_mutable_instances(self) -> Iterable[Mutable]:
        """
        Get any other :py:class:`betty.mutability.Mutable` instances contained by this one.
        """
        return ()

    @property
    def is_mutable(self) -> bool:
        """
        Whether the instance is mutable.
        """
        return self._mutable

    def mutable(self) -> None:
        """
        Mark the instance mutable.
        """
        self._mutable = True
        for instance in self.get_mutable_instances():
            instance.mutable()

    @property
    def is_immutable(self) -> bool:
        """
        Whether the instance is immutable.
        """
        return not self._mutable

    def immutable(self) -> None:
        """
        Mark the instance immutable.
        """
        self._mutable = False
        for instance in self.get_mutable_instances():
            instance.immutable()

    def assert_mutable(self) -> None:
        """
        Assert that the instance is mutable.

        :raise ImmutableError: if the instance is immutable.
        """
        if not self._mutable:
            raise ImmutableError(
                f"{self} was unexpectedly immutable, and cannot be modified."
            )

    def assert_immutable(self) -> None:
        """
        Assert that the instance is immutable.

        :raise MutableError: if the instance is mutable.
        """
        if self._mutable:
            raise MutableError(f"{self} was unexpectedly mutable, and can be modified.")


def mutable(*instances: Mutable) -> None:
    """
    Mark the given instances mutable.
    """
    for instance in instances:
        instance.mutable()


def immutable(*instances: Mutable) -> None:
    """
    Mark the given instances immutable.
    """
    for instance in instances:
        instance.immutable()
